#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/input.h>
#include <linux/poll.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>

struct cs_buffer {
  void *buffer;
  int size;
};

static struct cs_buffer csb;
static struct fasync_struct *fasync_queue = NULL;

#define COOLSOUND_NATIVE_MAJOR 244
static char sound_name[] = "tomtomgo-sound";

static int sound_open(struct inode *inode, struct file *fp)
{
       if (!(fp->f_mode & FMODE_WRITE || fp->f_mode & FMODE_READ))
               return -ENODEV;

       return 0;
}

static int sound_release(struct inode *inode, struct file *fp)
{
       if (!(fp->f_mode & FMODE_WRITE || fp->f_mode & FMODE_READ))
               return -ENODEV;
       return 0;
}

static void coolsound_vma_open(struct vm_area_struct *vma) {
}

static void coolsound_vma_close(struct vm_area_struct *vma) {
}

static int coolsound_native_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg) {
  return -ENOTTY;
}


static struct page *coolsound_vma_nopage(struct vm_area_struct *vma, unsigned long address, int *type) {
       unsigned long offset;
       struct cs_buffer *csb = vma->vm_private_data;
       struct page *page = NOPAGE_SIGBUS;
       
       offset = (address - vma->vm_start) + (vma->vm_pgoff << PAGE_SHIFT);
       
       
       if(offset > csb->size) {
               return NULL;
       }

       page = vmalloc_to_page(csb->buffer + offset);
       get_page(page);
       if(type) *type = VM_FAULT_MINOR;

       return page;
}


static struct vm_operations_struct coolsound_vm_ops = {
       .open    = coolsound_vma_open,
       .close   = coolsound_vma_close,
       .nopage  = coolsound_vma_nopage,
};


static int sound_mmap(struct file *fp, struct vm_area_struct *vma)
{
       vma->vm_ops                      = &coolsound_vm_ops;
       vma->vm_flags           |= VM_RESERVED;
       vma->vm_private_data = &csb;
       
       return 0;
}

static ssize_t sound_write(struct file *fp, const char __user *data, size_t count, loff_t *offset) {
  char buffer[128];
  int res, written = 0, i;

  while(count) {
    int maxwrite = sizeof(buffer);

    if(maxwrite > count) maxwrite = count;
    res = copy_from_user(buffer, data + written, maxwrite);
    if(res) return res;

    for(i=0; i<maxwrite; i+=2) {
      while((SASR0 & SASR0_TNF) == 0);
      SADR = buffer[i] | (buffer[i] << 16);
    }

    written += maxwrite;
    count -= maxwrite;
   }
  printk("Wrote %d bytes to audio device\n", written);
  return written;
}

static int sound_fasync(int fd, struct file *fp, int mode)
{
       return fasync_helper(fd, fp, mode, &fasync_queue);
}

static struct file_operations coolsound_native_fops = {
       .owner   = THIS_MODULE,
       .open    = sound_open,
       .release = sound_release,
       .ioctl   = coolsound_native_ioctl,
       .mmap    = sound_mmap,
       .fasync  = sound_fasync,
       .write   = sound_write,
};

static int __init cs_mod_init(void)
{
	register_chrdev(COOLSOUND_NATIVE_MAJOR, sound_name, &coolsound_native_fops);
	return 0;
}

static void __exit cs_mod_exit(void)
{
	unregister_chrdev(COOLSOUND_NATIVE_MAJOR, sound_name);
}

module_init(cs_mod_init);
module_exit(cs_mod_exit);

MODULE_AUTHOR("Rasmus Rohde");
MODULE_DESCRIPTION("TomTom CS Driver");
MODULE_LICENSE("GPL");
