#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

#include "hw.h"

static struct gotype gotype_current;

inline void IO_InitAberdeen(void)
{
        gotype_current.id                 = 5;
	gotype_current.name               = "TomTom GO 300";
	gotype_current.usbname            = "GO 300";
	gotype_current.familyname         = "TomTom GO";
	gotype_current.projectname        = "MALAGA";
	gotype_current.btname             = gotype_current.name;
	gotype_current.requiredbootloader = "2.02";
	gotype_current.cputype            = 0;
	gotype_current.linuxfreq          = 200000000;
	gotype_current.bootloaderfreq     = 200000000;

	gotype_current.bluetooth          = 0;
	gotype_current.btchip             = 0;
	gotype_current.btusb              = 0;
	gotype_current.btspeed            = 921600;
	gotype_current.btclock            = 16000000; /* all production aberdeens use 16 Mhz */
	gotype_current.btclass            = 0x001f00;
	gotype_current.handsfree          = 0;
	gotype_current.headsetgw          = 0;
	gotype_current.a2dp               = 0;

	gotype_current.lowbutton          = 1;
	gotype_current.battery            = 0;
	gotype_current.sdcard             = 1;
	gotype_current.tfttype            = 2;
	gotype_current.touchscreen        = 0;
	gotype_current.tsxchannel         = 5;
	gotype_current.tsychannel         = 7;
	gotype_current.gpstype            = 0;
	gotype_current.gpsephemeris       = 0;
	gotype_current.headphoneconnector = 0;
	gotype_current.speaker            = 0;
	gotype_current.codectype          = 0;
	gotype_current.usbslavetype       = 0;
	gotype_current.usbproductid       = 0;
	gotype_current.usbbuspowered      = 0; /* place R408 for USB charge function */
	gotype_current.memsize            = 32;
	gotype_current.ohciports          = 0;
	gotype_current.tftsoftstart       = 1;

	gotype_current.tsfets             = 0;
	gotype_current.acctype            = 0;
	gotype_current.gyro               = 0;
        gotype_current.remote             = 1;
}

static const char hw_proc_root_name[] = "barcelona";
static struct proc_dir_entry *hw_proc_root;
struct hw_proc_entry {
       const char *name;
       read_proc_t *func;
};

unsigned int reset_state;

static int hw_proc_read(char *page, char **start, off_t offset, int count, int *eof, 
const char *msg)
{
       int avail = strlen(msg) - offset;

       if (count >= avail) {
               count = avail;
               *eof = 1;
       }
       if (count <= 0)
               return 0;

       *start = page;
       memcpy(page, msg + offset, count);

       return count;
}


static int hw_proc_read_int(char *page, char **start, off_t offset, int count, int *eof, int i)
{
	unsigned char msg[16];
	snprintf(msg, sizeof msg, "%d\n", i);
	return hw_proc_read(page, start, offset, count, eof, msg);
}

static int hw_proc_read_str(char *page, char **start, off_t offset, int count, int *eof, const char *s)
{
	unsigned char msg[128];
	snprintf(msg, sizeof msg, "%s\n", s);
	return hw_proc_read(page, start, offset, count, eof, msg);
}

#define PROC_INT(name, func) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_int(page, start, offset, count, eof, IO_##func()); \
	}

#define PROC_INT_VARIABLE(name, var) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_int(page, start, offset, count, eof, var); \
	}

#define PROC_STR(name, func) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_str(page, start, offset, count, eof, IO_##func()); \
	}

PROC_INT(modelid, GetModelId)
PROC_STR(modelname, GetModelName)
PROC_STR(familyname, GetFamilyName)
PROC_STR(usbname, GetUsbName)
PROC_STR(btname, GetBtName)
PROC_INT(cputype, GetCpuType)
PROC_INT(cpufreq, GetCpuFreq)

PROC_INT(bluetooth, HaveBluetooth)
PROC_INT(btchip, GetBluetoothChip)
PROC_INT(btusb, HaveBluetoothUsb)
PROC_INT(btspeed, GetBluetoothSpeed)
PROC_INT(btclock, GetBluetoothClock)
PROC_INT(btclass, GetBluetoothClass)
PROC_INT(handsfree, HaveHandsfree)
PROC_INT(headsetgw, HaveHeadsetGw)
PROC_INT(a2dp, HaveA2DP)
PROC_INT(lowbutton, HaveLoweredButton)
PROC_INT(battery, HaveBattery)
PROC_INT(sdcard, HaveSdCard)
PROC_INT(harddisk, HaveHarddisk)
PROC_INT(tsfets, HaveTsFets)
PROC_INT(remote, HaveRemote)
PROC_INT(docking, HaveDocking)
PROC_INT(tfttype, GetTftType)
PROC_INT(gpstype, GetGpsType)
PROC_INT(gpsephemeris, HaveGpsEphemeris)
PROC_INT(headphoneconnector, HaveHeadphoneConnector)
PROC_INT(codectype, GetCodecType)
PROC_INT(codecmaster, GetCodecMaster)
PROC_INT(acctype, GetAccType)
PROC_INT(usbslavetype, GetUSBSlaveType)
PROC_INT(lightsensor, HaveLightSensor)
PROC_INT(gyro, HaveGyro)
PROC_INT(loquendo, HaveLoquendo )
PROC_INT_VARIABLE(resetstate, reset_state )

#define HW_PROC_ENTRY(x) { .name = #x, .func = hw_proc_##x }

static struct hw_proc_entry hw_proc_entries[] = {
	HW_PROC_ENTRY(modelid),
	HW_PROC_ENTRY(modelname),
	HW_PROC_ENTRY(familyname),
	HW_PROC_ENTRY(usbname),
	HW_PROC_ENTRY(btname),
	HW_PROC_ENTRY(cputype),
	HW_PROC_ENTRY(cpufreq),
	HW_PROC_ENTRY(bluetooth),
	HW_PROC_ENTRY(btchip),
	HW_PROC_ENTRY(btusb),
	HW_PROC_ENTRY(btspeed),
	HW_PROC_ENTRY(btclock),
	HW_PROC_ENTRY(btclass),
	HW_PROC_ENTRY(handsfree),
	HW_PROC_ENTRY(headsetgw),
	HW_PROC_ENTRY(a2dp),
	HW_PROC_ENTRY(lowbutton),
	HW_PROC_ENTRY(battery),
	HW_PROC_ENTRY(sdcard),
	HW_PROC_ENTRY(harddisk),
	HW_PROC_ENTRY(tsfets),
	HW_PROC_ENTRY(remote),
	HW_PROC_ENTRY(docking),
	HW_PROC_ENTRY(tfttype),
	HW_PROC_ENTRY(gpstype),
	HW_PROC_ENTRY(gpsephemeris),
	HW_PROC_ENTRY(headphoneconnector),
	HW_PROC_ENTRY(codectype),
	HW_PROC_ENTRY(codecmaster),
	HW_PROC_ENTRY(acctype),
	HW_PROC_ENTRY(usbslavetype),
	HW_PROC_ENTRY(lightsensor),
	HW_PROC_ENTRY(gyro),
	HW_PROC_ENTRY(loquendo),
	HW_PROC_ENTRY(resetstate),
};

static int __init hw_mod_init(void)
{
	int i;

	IO_InitAberdeen();

	hw_proc_root = proc_mkdir(hw_proc_root_name, NULL);
	if (hw_proc_root == NULL) {
		printk("Unable to create hw /proc root %s\n", hw_proc_root_name);
		return -1;
	}

	for (i = 0; i < ARRAY_SIZE(hw_proc_entries); ++i) {
		if (create_proc_read_entry(hw_proc_entries[i].name, S_IRUGO, hw_proc_root, hw_proc_entries[i].func, 0) == NULL) {
			printk("Unable to create hw /proc entry %s\n", hw_proc_entries[i].name);
			return -1;
		}
	}

	return 0;
}

static void __exit hw_mod_exit(void)
{

}

module_init(hw_mod_init);
module_exit(hw_mod_exit);

MODULE_AUTHOR("Rasmus Rohde");
MODULE_DESCRIPTION("TomTom HW Driver");
MODULE_LICENSE("GPL");
